/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */
/************************************************************************* 
**************************************************************************
**                                                                      **
** this file includes routines that dump the current internal line      **
**                                                                      **
************************************************************************** 
*************************************************************************/ 

#include <stdio.h>
#include "TOKEN.h"
#include "lex.h"
#include "macros.h"

#define  max(a,b)       ((a<b) ? b : a)
#define  MAXZWC         100        /* maximum number of respective zero
              						width characters        */
TOKENPTR new_token();              /* function return type    */


extern   bool arabic_fonts[256];

/* new stretch flags for change #1 - harry */

extern   char stretch_mode;
extern   char stretch_place;
extern   float stretch_amount;
extern 	int msc_flag;
extern 	int msl_flag;
extern 	int msw_flag;  /* change #2 */

/************************************************************************* 
************************************************************************** 
**                                                                      **
** this routine dumps the current internal line while reversing         ** 
** the tokens of the specified lr direction                             **
** Was enhanced in order to deal with ZERO width characters, or         **
** ZERO horizontal movements (via the usage of '\z' for example).       **
** It was also optimized and made more efficient.                       **
**                                                                      **
** Johny                                                                **
************************************************************************** 
*************************************************************************/

dump_line(start,end,reverse_lr)

  TOKENPTR *start;
  TOKENPTR *end;
  bool     reverse_lr;

{

  bool     start_of_line = TRUE;

  TOKENPTR tmptr;
  TOKENPTR new_start = NULL;      /* ptr to start of new token line */
  TOKENPTR new_end   = NULL;      /* ptr to end of new token line   */
  TOKENPTR rev_start = NULL;      /* ptr to start of reversed token line */
  TOKENPTR rev_end   = NULL;      /* ptr to end of reversed token line   */
  TOKENPTR tokenptr  = *start;    /* next token to dump                  */
  TOKENPTR zerowc[MAXZWC];
  int      total_stretch = 0;
  int      i,j,hpos;
  
  while(tokenptr != NULL)
  {
    if  (tokenptr->lr != reverse_lr) 
    {                             /* if the language is written in it's
                                     natural direction                  */
      if (tokenptr->next == NULL)
        tokenptr->ending = NOT_END;
      simple_add_token(tokenptr, &new_start, &new_end);/* change #1 - */
						       /* harry */
      tokenptr = tokenptr->next;
    }
    else
    {
      while ((tokenptr != NULL) && (tokenptr->lr == reverse_lr))
      {
        if (tokenptr->ending) {            /* remark word endings */ 
            tokenptr->ending = NOT_END;
            tokenptr->begining = BEGINING;
        }
        else if (tokenptr->begining) {
            tokenptr->ending = END;
            tokenptr->begining = NOT_BEGIN;
        }
        hpos = tokenptr->horizontal_pos;
        i = 0;
        while ((tokenptr->next != NULL) &&
               (tokenptr->next->horizontal_pos == hpos))
        {
            zerowc[i++] = tokenptr;
            tokenptr = tokenptr->next;
        }
        tmptr = tokenptr->next;
        if (i>0)        /* a ZERO-width character was detected */
        {
          int temp;
    
          zerowc[i] = tokenptr;
          for (j=i; j>=0; j--) {
				zerowc[j]->width =
				  tokenBasicWidth(tokenptr); /*change #1 - harry */
            push_token(zerowc[j],&rev_start,&rev_end);
          }
          temp = zerowc[i]->ending;
          zerowc[i]->ending = zerowc[0]->ending;
          zerowc[0]->ending = NOT_END;
          zerowc[0]->begining = temp;
        }
        else push_token(tokenptr,&rev_start,&rev_end);
        tokenptr = tmptr;
      }

      recalculate_horizontal(rev_start,rev_end);

      if (tokenptr == NULL)            /* unmark word end at end of line */
        rev_end->ending = NOT_END;

      tmptr = rev_start;      /* dump reversed line */
      while (tmptr != NULL)
      {
	      simple_add_token(tmptr, &new_start, &new_end);/* change #1 - */
	                                                    /* harry */
	      tmptr = tmptr->next;
      }

     free_line(&rev_start,&rev_end); 
    }
  }

  /* Change #1 - Harry */

  /* stretch Arabic words and reposition letters .. Johny */
  total_stretch = calculate_total_stretching(new_start,new_end);
  if ((total_stretch < 0) ||      /* the line was shrinked!! */
      (number_of_stretch_candidates(new_start,new_end) == 0)) 
                                  /* nothing to stretch! */
    total_stretch = 0;
  stretch_line(new_start,new_end,total_stretch);
  
  tokenptr = new_start;
  while(tokenptr != NULL)
  {
	  put_token(tokenptr,start_of_line);
	  start_of_line = FALSE;
	  tokenptr = tokenptr->next;
  }

  free_line(start,end);
  free_line(&new_start,&new_end);
  
  return;
}


/*************************************************************************
**************************************************************************
**                                                                      **
** this routine reverses the specified tokens in-place in a line        **
** It also recognizes ZERO width characters and preserves their         **
** order with the next letter. (Change #2)                              **
**                                                                      **
**************************************************************************
*************************************************************************/


reverse_tokens(rev_start,rev_end)
TOKENPTR *rev_start;
TOKENPTR *rev_end;
{
	TOKENPTR tmp_start = NULL;
	TOKENPTR tmp_end = NULL;
	TOKENPTR tmptr;
	TOKENPTR tokenptr;
	TOKENPTR zerowc[MAXZWC];
	int i,j,hpos;

	tokenptr = *rev_start;
	while (tokenptr != NULL)
	  {
		  if (tokenptr->ending) {            /* remark word endings */ 
			  tokenptr->ending = NOT_END;
			  tokenptr->begining = BEGINING;
		  }
		  else if (tokenptr->begining) {
			  tokenptr->ending = END;
			  tokenptr->begining = NOT_BEGIN;
		  }
		  hpos = tokenptr->horizontal_pos;
		  i = 0;
		  while ((tokenptr->next != NULL) &&
			 (tokenptr->next->horizontal_pos == hpos))
		    {
			    zerowc[i++] = tokenptr;
			    tokenptr = tokenptr->next;
		    }
		  tmptr = tokenptr->next;
		  if (i>0)        /* a ZERO-width character was detected */
		    {
			    int temp;
			    
			    zerowc[i] = tokenptr;
			    for (j=i; j>=0; j--) {
				    zerowc[j]->width =
				  tokenBasicWidth(tokenptr); /*change #1 - harry */
				    push_token(zerowc[j],&tmp_start,&tmp_end);
			    }
			    temp = zerowc[i]->ending;
			    zerowc[i]->ending = zerowc[0]->ending;
			    zerowc[0]->ending = NOT_END;
			    zerowc[0]->begining = temp;
		    }
		  else push_token(tokenptr,&tmp_start,&tmp_end);
		  tokenptr = tmptr;
	  }
	
	if (tokenptr == NULL)            /* unmark word end at end of line */
	  tmp_end->ending = NOT_END;

	*rev_start = tmp_start;
	*rev_end   = tmp_end;
	
	recalculate_horizontal(*rev_start,*rev_end);
}


/*************************************************************************
**************************************************************************
**                                                                      **
** this routine reverses the specified internal line                    **
** It also recognizes ZERO width characters and preserves their         **
** order with the next letter.                                          **
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/


reverse_line(start,end,paper_width)

TOKENPTR *start;
TOKENPTR *end;
int      paper_width;

{
  int       line_length = (*end)->horizontal_pos +
                   tokenBasicWidth(*end); /*change #1 - harry */
  int       new_indent  = paper_width - line_length - 1;
  int       adjustment  = new_indent - (*start)->horizontal_pos;
  TOKENPTR  tmptr = (*start);
  TOKENPTR  save_start;
  TOKENPTR  tmp_start   = NULL;
  TOKENPTR  tmp_end     = NULL;
  bool      begining;
  bool      ending;
  TOKENPTR  zerowc[MAXZWC];
  int       i,j,hpos;

  (*end)->ending = END;

  while ((tmptr = (*start)) != NULL)
  {
    *start = (*start)->next;

    begining = tmptr->begining;
    ending   = tmptr->ending;
    tmptr->begining = ending;
    tmptr->ending   = begining;
    hpos = tmptr->horizontal_pos;
    i = 0;
	/* enter the equaly hpos tokens into an array */
    while ((tmptr->next != NULL) && (tmptr->next->horizontal_pos == hpos))
    {
        zerowc[i++] = tmptr;
        tmptr = tmptr->next;
    }
    if (i>0) {
        int temp;

		  zerowc[i] = tmptr;
        for (j=i-1; j>=0; j--)		/* update the start pointer */
            *start = (*start)->next;
        save_start = *start;
		/* update the width for each one of them */
        for (j=i; j>=0; j--) {
            zerowc[j]->width = tokenBasicWidth(tmptr);
	                             /* take the last token */
                                     /* width */
                         	      /*change #1 - harry */

            zerowc[j]->horizontal_pos += adjustment;
            push_token(zerowc[j],&tmp_start,&tmp_end);
        }
        temp = zerowc[i]->ending;
        zerowc[i]->ending = zerowc[0]->ending;
        zerowc[0]->ending = NOT_END;
        zerowc[0]->begining = temp;
        *start = save_start;
    }
    else if (tmptr != NULL) {
        tmptr->horizontal_pos += adjustment;
        push_token(tmptr,&tmp_start,&tmp_end);
    }
  }

  *start = tmp_start;
  *end   = tmp_end;

  recalculate_horizontal(*start,*end);
  
  return;
}

/************************************************************************* 
************************************************************************** 
**                                                                      **
** this routine recalculates the horizontal motion in the specified     **
** internal line                                                        **
**                                                                      **
************************************************************************** 
*************************************************************************/ 

recalculate_horizontal(start,end)

  TOKENPTR  start;                  /* ptr to start of line */
  TOKENPTR  end  ;                  /* ptr to end of line   */

{

  TOKENPTR prev_token;
  TOKENPTR next_token;
  int      new_horizontal;
  int      prev_horizontal;
  int      start_horizontal;

  start_horizontal = start->horizontal_pos;
  prev_token = start; 
  prev_horizontal = end->horizontal_pos;
  new_horizontal  = prev_horizontal;
  
  if(prev_token->next != NULL)
  {
     next_token = prev_token->next;
     while (next_token != NULL)
     {
	     /*change #1 - harry */

       new_horizontal = new_horizontal + (prev_token->horizontal_pos -
                        next_token->horizontal_pos) +
                        tokenBasicWidth(prev_token)
			  - tokenBasicWidth(next_token);
       prev_token->horizontal_pos = prev_horizontal;
       prev_token = next_token;
       prev_horizontal = new_horizontal;
       next_token = prev_token->next;
     }
   }
   prev_token->horizontal_pos = new_horizontal;
 }

/*************************************************************************
**************************************************************************
**                                                                      **
** This function should return the total stretching amount.             **
** It does that, by calculating the total amount of extra interword     **
** spaces.                                                              **
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/

int
calculate_total_stretching (start,end)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */

{
  TOKENPTR prev_token;
  TOKENPTR next_token;
  TOKENPTR tokenptr  = start;
  int      total_stretch = 0;
  int      space_width;

  if ((tokenptr != NULL) && (tokenptr->next != NULL))
  {
		prev_token = tokenptr;
		next_token = tokenptr->next;
		while (next_token != NULL)
		{
		  /* we have to calculate the space width for each word, as the
			  point size can change at any point */
			/* rounding error fix. change #1 - harry */
			
		  space_width = width1(0, prev_token->point_size,
				       prev_token->font);
			     
		  if (prev_token->ending) {
			  /*change #1 - harry */

			  total_stretch = total_stretch + 
			    (next_token->horizontal_pos -
			     prev_token->horizontal_pos) -
			       (tokenBasicWidth(prev_token)
				+ space_width);
		  }
		  prev_token = next_token;
		  next_token = prev_token->next;
		}
      tokenptr = next_token;
  }

  return(total_stretch);
}


/* All the following functions were changed for change #1 - harry */
/* Some of them are based on previous functions and some are completly */
/* new */



/*************************************************************************
**************************************************************************
**                                                                      **
** stretch a line by 'total_stretch' according to the current stretch   **
** mode and place.                                                      **
**                                                                      **
**************************************************************************
*************************************************************************/

stretch_line (start,end,total_stretch)
TOKENPTR start;           	/* ptr to start of line */
TOKENPTR end;              /* ptr to end of line   */
int total_stretch;
{
	int remainder = 0;
	int candidates;

	if ((total_stretch == 0) || (stretch_mode == 'n'))
		return;

	switch (stretch_place) {
	case 'f':
	case '2':
	case 'm':
		{
			bool finished = FALSE;
			TOKENPTR tokenptr = start;
			bool letter_stretch;
			bool connect_stretch;
			TOKENPTR token;
			
			while ((tokenptr != NULL) && !finished) {
				if (candidates = word_stretch_candidates(tokenptr,
							    &letter_stretch, 
							    &connect_stretch, & token)) {
					if (stretch_place == '2') 
					  remainder = stretch_word(tokenptr, 
								   total_stretch/candidates, TRUE);
					else remainder = stretch_word(tokenptr, 
								      total_stretch, TRUE);
					if (stretch_place == 'm')
					  total_stretch = remainder;
					if ((stretch_place != 'm') || (remainder == 0)) {
						recalculate_spaces(start,end,remainder);
						finished = TRUE;
					}
				}
				while ((tokenptr != NULL) && !(tokenptr->ending))
				  tokenptr = tokenptr->next;
				if (tokenptr != NULL)
				  tokenptr = tokenptr->next;
				if (!finished && (tokenptr == NULL) &&
				    (stretch_place == 'm')) {
					recalculate_spaces(start,end,remainder);
					finished = TRUE;
				}
			}
			break;
		}

	case 'a':
	case 'd':
		remainder = spread_stretch_in_line(start, end, total_stretch);
		recalculate_spaces(start,end,remainder);
		break;
	};
	return;
}


/*************************************************************************
**************************************************************************
**                                                                      **
** recalculates the inter words spaces, after stretching a specif-      **
** ied letter. It simply removes the  extra spaces added by  troff      **
** and spread the amount of remainder, which is the amount of poi-      **
** nts less that the filler width.                                      **
** initialization: hmove = 0                                            **
** when reaching an end of word: hmove = hmove - extra_space            **
** when reaching a letter with a filler: hmove = hmove + filler         **
**                                                                      **
** Johny.      											  **
** Modified to handle letter stretching as well.                        **
**                                                                      **
**************************************************************************
*************************************************************************/

recalculate_spaces (start,end,remainder)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
int       remainder;

{
  TOKENPTR prev_token;
  TOKENPTR next_token;
  TOKENPTR tokenptr  = start;
  int      space_width=0, filler_width=0;
  int      extra_space=0, hmove=0;
  int      prev_horizontal;
  int      nw, rpw=0;                 /* remainder per word   */
  bool     last_word = TRUE;

  if ((nw = number_of_words(start,end)) == 0)
	 return;
  if (nw == 1)
	rpw = 0;
  else rpw = (int) (remainder / (nw - 1));

  if ((tokenptr != NULL) && (tokenptr->next != NULL))
  {
      prev_token = tokenptr;
      next_token = tokenptr->next;
      prev_horizontal = prev_token->horizontal_pos;
      while (next_token != NULL)
		{
		  space_width = width1(0, prev_token->point_size,
				       prev_token->font);
		  filler_width = width2("hy",
					next_token->point_size,
					next_token->font);
		  if (prev_token->ending) {
			  extra_space = (next_token->horizontal_pos -
					 prev_horizontal) -
					   (tokenFullWidth(prev_token)
					    + space_width);
			  hmove = hmove - extra_space + rpw;
			  if (last_word) {
				  if (nw > 1)
				    hmove = hmove + (remainder % nw);
				  last_word = FALSE;
			  }
   } else
     {
	     if (prev_token->fillers_num > 0)
	       hmove += prev_token->fillers_num * filler_width;
	     if (prev_token->stretch_filler_width > 0)    /* change #2 */
	       hmove += prev_token->stretch_filler_width;
	     if (prev_token->stretch_width > 0)
	       hmove += prev_token->stretch_width;
	     if (prev_token->AC_width > 0)                /* change #2 */
	       hmove += prev_token->AC_width;
	     if (prev_token->BC_width > 0)                /* change #2 */
	       hmove += prev_token->BC_width;
     }
		  prev_horizontal = next_token->horizontal_pos;
		  next_token->horizontal_pos += hmove;
		  prev_token = next_token;
		  next_token = next_token->next;
	  }
      tokenptr = next_token;
  }
}


/*************************************************************************
**************************************************************************
**                                                                      **
** return the number of words in a line.                                **
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

int number_of_words (start,end)
TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
{
  TOKENPTR  tokenptr = start;
  int       nw = 0;               /* the number of words in line */

  while (tokenptr != NULL) {
    if (tokenptr->ending)
        nw++;
    else if (tokenptr->next == NULL)
      nw++;
    tokenptr = tokenptr->next;
  }
  return(nw);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** return if token has a manually stretched letter (change #2)          **
**                                                                      **
**************************************************************************
*************************************************************************/

bool manually_stretched_letter (token)
TOKENPTR token;               /* token to be check */
{
     return (token->manual_stretch_width > 0);
}


/*************************************************************************
**************************************************************************
**                                                                      **
** return if token has a manually stretched connection (change #2)      **
**                                                                      **
**************************************************************************
*************************************************************************/

bool manually_stretched_connection (token)
TOKENPTR token;               /* token to be check */
{
	if ((token->next != NULL) &&
		(token->next->token_type == C_token) &&
		(token->next->char1 == 'h') &&
		(token->next->char2 == 'y'))
          return 1;
     return ((token->manual_BC_width > 0) ||
               (token->manual_AC_width > 0));
}


/*************************************************************************
**************************************************************************
**                                                                      **
** return the number of stretch candidates in a word according to the   **
** current stretch mode and a pointer to a token that is to stretched   **
** including flags indicating if the letter or connection are to be     **
** stretched.                                                           **
**                                                                      **
**************************************************************************
*************************************************************************/

int word_stretch_candidates (start, letter_stretch_flag,
			     connect_stretch_flag, token)
TOKENPTR  start;               /* ptr to start of word */
bool* letter_stretch_flag;     /* token should be letter stretched */
bool* connect_stretch_flag;    /* token should be connection stretched */
TOKENPTR* token;               /* token to be stretched if found */
{
	TOKENPTR tokenptr = start;
	bool finished = FALSE;
	char letter[4];
	int i;
	int candidates = 0;
	int stretchable;         /* change #2 */
	int connectable;


	/* change #2 - harry */
	
     if (msw_flag) {
     	while (tokenptr != NULL) {
     		if ((arabic_fonts[tokenptr->font] == ARABIC) &&
			    (manually_stretched_letter(tokenptr) ||
			     manually_stretched_connection(tokenptr)))
			  return 0;
     		if (tokenptr->ending)
     		  tokenptr = NULL;
     		else tokenptr = tokenptr->next;
          }
	}
		
	tokenptr = start;
	*letter_stretch_flag = FALSE;
	*connect_stretch_flag = FALSE;
	while ((tokenptr != NULL) && !finished) {
		if (arabic_fonts[tokenptr->font] == ARABIC) {
			/* change #2 - harry */

			if (tokenptr->token_type == c_token) {
				stretchable = stretchable1(tokenptr->char1, tokenptr->font);
				connectable = connectable1(tokenptr->char1, tokenptr->font);
			}
			if (tokenptr->token_type == C_token) {
				sprintf(letter, "%c%c",tokenptr->char1,tokenptr->char2);
				stretchable = stretchable2(letter, tokenptr->font);
				connectable = connectable2(letter, tokenptr->font);
			}
			if (tokenptr->token_type == N_token) {
				sprintf(letter,"%c%c%c",tokenptr->char1,tokenptr->char2,
					tokenptr->char3);
				i = atoi(letter);
				stretchable = stretchablen(i, tokenptr->font);
				connectable = connectablen(i, tokenptr->font);
			}

			stretchable = stretchable && 
			  font_letter_stretchable(tokenptr->font);
			connectable = connectable && 
			  font_connection_stretchable(tokenptr->font);

			if ((stretch_mode == 'l') || (stretch_mode == 'e')) {
				if (stretchable && 
				    !(msl_flag && 
				      manually_stretched_letter(tokenptr))) {
					*letter_stretch_flag = TRUE;
					*token = tokenptr;
					finished = TRUE;
					candidates = 1;
				}
			}
			if (((stretch_mode == 'c') || (stretch_mode == 'e'))
			    && !finished) {
				if (connectable &&
				    !(msc_flag && 
				      manually_stretched_connection(tokenptr))) {
					*connect_stretch_flag = TRUE;
					*token = tokenptr;
					finished = TRUE;
					candidates = 1;
				}
			}
			if (stretch_mode == 'b') {
				if (stretchable && 
				    !(msl_flag && 
				      manually_stretched_letter(tokenptr)) &&
				    !(msc_flag && 
				      manually_stretched_connection(tokenptr))) {
					*letter_stretch_flag = TRUE;
					*token = tokenptr;
					finished = TRUE;
					candidates = 1;
				}
				if (connectable &&
                                    !(msl_flag &&
                                      manually_stretched_letter(tokenptr)) &&
                                    !(msc_flag &&
                                      manually_stretched_connection(tokenptr))) {
					*connect_stretch_flag = TRUE;
					*token = tokenptr;
					finished = TRUE;
					candidates += 1;
				}
			}
		}
		if (tokenptr->ending)
		  finished = TRUE;
		else tokenptr = tokenptr->next;
	}

	return candidates;
}


/*************************************************************************
**************************************************************************
**                                                                      **
** return the number of stretch candidates in a line according to the   **
** current stretch mode.                                                **
**                                                                      **
**************************************************************************
*************************************************************************/

int number_of_stretch_candidates (start,end)
TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
{
	TOKENPTR  tokenptr = start;
	bool letter_stretch;
	bool connect_stretch;
	TOKENPTR token;
	int       nsc = 0;        /* the number of stretch candidates in line */

	while (tokenptr != NULL) {
		nsc += word_stretch_candidates(tokenptr,
					       &letter_stretch,
					       &connect_stretch, &token);
		while ((tokenptr != NULL) && !(tokenptr->ending))
		  tokenptr = tokenptr->next;
		if (tokenptr != NULL)
		  tokenptr = tokenptr->next;
	}

	return nsc;
}


/*************************************************************************
**************************************************************************
**                                                                      **
** stretches a letter by 'stretch' and returns the remainder if any.    **
** if the 'do_stretch' flag is false no actual stretching is preformed. **
**                                                                      **
**************************************************************************
*************************************************************************/

int letter_stretch (tokenptr, stretch, do_stretch)
TOKENPTR tokenptr;
int stretch;
bool do_stretch;
{
	if (do_stretch)
		tokenptr->stretch_width += stretch;
	return 0;
}


/*************************************************************************
**************************************************************************
**                                                                      **
** stretches a connection to a letter by 'stretch' and returns the      **
** remainder if any.                                                    **
** if the 'do_stretch' flag is false no actual stretching is preformed. **
** Updated in change #2.                                                **
**                                                                      **
**************************************************************************
*************************************************************************/

int connect_stretch (tokenptr, stretch, do_stretch)
TOKENPTR tokenptr;
int stretch;
bool do_stretch;
{
	int filler_width;
	int connect_after;
	char letter[4];
	int i;

	switch (font_connection_stretch(tokenptr->font)) {
	      case FIXED_FILLER:
		{
			filler_width = width2("hy", tokenptr->point_size, tokenptr->font);
			if (do_stretch) {
				tokenptr->filler_width = filler_width;
				tokenptr->fillers_num += (int) (stretch / filler_width);
			}
			
			return (stretch % filler_width);
		}
	      case STRETCHABLE_FILLER:
		{
			if (do_stretch)
			  tokenptr->stretch_filler_width += stretch;
			return 0;
		}
	      case STRETCHABLE_CONNECTIONS:
		{
			if (do_stretch) {
				connect_after = 0;
				if (tokenptr->next != NULL) {
					if (tokenptr->next->token_type == c_token) {
						if (connectAfter1(tokenptr->next->char1, 
								  tokenptr->next->font))
						  connect_after = 1;
					} 
					else if (tokenptr->next->token_type == C_token) {
						sprintf(letter,
							"%c%c",tokenptr->next->char1,
							tokenptr->next->char2);
						if (connectAfter2(letter, tokenptr->next->font))
						  connect_after = 1;
					} 
					else if (tokenptr->next->token_type == N_token) {
						sprintf(letter,"%c%c%c",
							tokenptr->next->char1,
							tokenptr->next->char2,
							tokenptr->next->char3);
						i = atoi(letter);
						if (connectAftern(i, tokenptr->next->font))
						  connect_after = 1;
					}			
				}
	
				if (connect_after) {
					tokenptr->BC_width += stretch / 2;
					tokenptr->next->AC_width += 
					  (stretch / 2) + (stretch % 2);
				}
				else tokenptr->BC_width += stretch;
			}
			return 0;
		}
	      default:
		return stretch;
	};
}


/*************************************************************************
**************************************************************************
**                                                                      **
** stretches a word by 'stretch' according to the current stretch mode  **
** and place and returns the remainder if any.                          **
** if the 'do_stretch' flag is false no actual stretching is preformed. **
**                                                                      **
**************************************************************************
*************************************************************************/

int stretch_word(start, stretch, do_stretch)
TOKENPTR start;
int stretch;
bool do_stretch;
{
	bool letter_stretch_flag;
	bool connect_stretch_flag;
	TOKENPTR tokenptr;
	int candidates;
	int remainder;
	int temp;
	int st;
	int st_amount;     /* stretch amount in emms in 'm' mode */
	
	if (stretch == 0)
	  return 0;

	candidates = word_stretch_candidates(start, &letter_stretch_flag,
					     &connect_stretch_flag, &tokenptr);
	
	if (candidates == 0)   /* can't stretch this word */
	  return stretch;

	st_amount = (stretch_amount * 6 * width2("\\|", tokenptr->point_size, 
						 tokenptr->font));

	if (stretch_place == 'f') {  /* make sure only one candidate */
				     /* is stretched */
		if (letter_stretch_flag == TRUE)
		  connect_stretch_flag = FALSE;
	}

	remainder = 0;
	if (connect_stretch_flag) {
		if (stretch_place == 'm') {
			if (stretch > st_amount)
			  st = st_amount;
			else st = stretch;
			remainder = stretch+connect_stretch(tokenptr,st,do_stretch)-st;
		}
		else remainder = connect_stretch(tokenptr,stretch,do_stretch);
		if (stretch_place == 'm')
		  stretch = remainder;
		else stretch += remainder;
	}
	if (letter_stretch_flag) {
		if (stretch_place == 'm') {
			if (stretch > st_amount)
			  st = st_amount;
			else st = stretch;
			remainder = stretch+letter_stretch(tokenptr,st,do_stretch)-st;
		}
		else remainder = letter_stretch(tokenptr, stretch, do_stretch);
	}

	return remainder;
}


/*************************************************************************
**************************************************************************
**                                                                      **
** returns the total remainders from all of the stretchable candidates  **
** assuming they are stretched with stretch 'spc'                       **
**                                                                      **
**************************************************************************
*************************************************************************/

int stretch_remainders (start,end,spc)

TOKENPTR start; 	     /* ptr to start of line */
TOKENPTR end ;           /* ptr to end of line   */
int spc;				/* stretch per candidate */
{
	TOKENPTR  tokenptr = start;
	int remainder = 0;
	bool letter_stretch_flag;
	bool connect_stretch_flag;
	TOKENPTR temp;

	while (tokenptr != NULL) {
		if (word_stretch_candidates(tokenptr,
					    &letter_stretch_flag,
					    &connect_stretch_flag,
					    &temp) > 0)
		  remainder += stretch_word(tokenptr, spc, FALSE);
		while ((tokenptr != NULL) && !(tokenptr->ending))
		  tokenptr = tokenptr->next;
		if (tokenptr != NULL)
		  tokenptr = tokenptr->next;
	}
	
	return remainder;
}

/*************************************************************************
**************************************************************************
**                                                                      **
** stretches all the stretchable units in the line according to the     **
** stretch mode and place by dividing the total stretch evenly among    **
** them.                                                                **
**                                                                      **
**************************************************************************
*************************************************************************/

int spread_stretch_in_line (start,end,total_stretch)
TOKENPTR start;            /* ptr to start of line */
TOKENPTR end  ;            /* ptr to end of line   */
int total_stretch;      	/* total stretch to spread in line */
{
	int nc = 0;					/* number of candidates */
	int spc = 0;					/* stretch per candidate */
	int last_word_remainder;
	bool last_word = TRUE;
	TOKENPTR tokenptr = start;
	int candidates;
	int remainder;
	int letter_candidates = 0;    /* letter candidates in 'a' */
          				     /* stretch place */
	int splc = 0;       /* stretch per letter candidate in 'a' */
			          /* place */
	int splc_extra = 0;    /* stretch per letter candidate in 'a' */
          			   /* place extra only in some words */
	char temp;
	bool letter_stretch;
	bool connect_stretch;
	TOKENPTR token;

	if ((nc = number_of_stretch_candidates(start, end)) == 0)
		return total_stretch;

	spc = (int) (total_stretch / nc);

	remainder = (total_stretch % nc) +
	  stretch_remainders(start,end, spc);

	if ((stretch_place == 'a') && (remainder > 0) &&
		    (stretch_mode != 'c')) {
			temp = stretch_mode;
			stretch_mode = 'l';
			letter_candidates = 
			  number_of_stretch_candidates(start, end);
			stretch_mode = temp;
			if (letter_candidates > 0) {
			  splc = (int) (remainder / letter_candidates);
			  if (remainder % letter_candidates > 0)
			    splc_extra = 1;
		  }
		}

	while (tokenptr != NULL) {
		candidates = word_stretch_candidates(tokenptr,
						     &letter_stretch,
						     &connect_stretch,
						     &token);
		if (candidates > 0) {
			stretch_word(tokenptr, spc, TRUE);
			if (letter_candidates > 0) {
				temp = stretch_mode;
				stretch_mode = 'l';
				candidates = word_stretch_candidates(tokenptr,
								     &letter_stretch,
								     &connect_stretch,
								     &token);
				if ((candidates > 0) && (remainder > 0)) {
					stretch_word(tokenptr,
						     splc + splc_extra, TRUE);
					remainder -= splc*candidates + splc_extra;
				}
				stretch_mode = temp;
			}
		}
		while ((tokenptr != NULL) && !(tokenptr->ending))
		  tokenptr = tokenptr->next;
		if (tokenptr != NULL)
		  tokenptr = tokenptr->next;
	}

	return remainder;
}


/* used for debugging ... Johny */

print_line(start,end)
TOKENPTR  start;                 /* ptr to start of line */
TOKENPTR  end  ;                 /* ptr to end of line   */
{

  TOKENPTR tmptr = start;
  
  while (tmptr != NULL) {
	 printf("horiz:%d , vert:%d , manual stretch:%d, stretch:%d fillers:%d char:%c%c%c , width:%d, start:%d, end:%d\n",
				tmptr->horizontal_pos,tmptr->vertical_pos,
				tmptr->manual_stretch_width, tmptr->stretch_width,
				tmptr->fillers_num,tmptr->char1,
			tmptr->char2,tmptr->char3,tmptr->width,
		tmptr->begining, tmptr->ending);
    tmptr = tmptr->next;
  }
  printf("\n");
}



