/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */
/*********************************************************
**********************************************************
**                                                      **
** this file contains a number of routines that work on **
** internal tokens and token lines                      **
**                                                      **
**********************************************************
*********************************************************/

#include  <stdio.h>
#include  <math.h>       /* Change #2 */
#include  "TOKEN.h"
#include  "TABLE.h"
#include  "macros.h"
#include  "lex.h"



extern int  out_font;                  /* current output font */
extern int  out_size;                  /* current output size */
extern int  out_horizontal;            /* current output horizontal position */
extern int  out_vertical;              /* current output vertical position */
extern char out_font_name[];           /* current output font name */
extern TABLENTRY out_fontable[];       /* current output font table */
extern int  in_font;                   /* current input font */
extern int  in_size;                   /* current input size */
extern int  in_horizontal;             /* current input horizontal position */
extern int  in_vertical;               /* current input vertical position */
extern char in_font_name[];            /* current input font name */
extern TABLENTRY in_fontable[];        /* currnet input font table */
extern bool in_lr;                     /* current input font direction */
extern bool direction_table[];         /* direction of font i */

/* Change #2 */

extern float font_slant();             /* function return type */

/****************************************************************
*****************************************************************
**                                                             **
** this routine allocates and initializes a new internal token **
**                                                             **
*****************************************************************
****************************************************************/

TOKENPTR
new_token (toktyp,ptsize,font,horizontal,vertical,fontdir,beg_wrd,end_wrd,char1,
           char2,char3,f_name,width,manual_stretch,manual_AC_stretch,
	   manual_BC_stretch)

  int          toktyp;         /* token type                 */
  int          ptsize;         /* point size                 */
  int          font;           /* font number                */
  int          horizontal;     /* horizontal position        */
  int          vertical;       /* vertical position          */
  bool         fontdir;        /* font direction             */
  bool         beg_wrd;        /* begining of word indicator */
  bool         end_wrd;        /* end of word indicator      */
  char         char1;          /* 1st char of (abs) character token*/
  char         char2;          /* 2nd char of (abs) character token*/
  char         char3;          /* 3nd char of abs character token*/
  char         *f_name;        /* font name                  */
  int          width;          /* character width            */
  int          manual_stretch; /* manual letter stretch. change #1 - harry */
  int          manual_AC_stretch; /* manual AC stretch. change #2 - harry */
  int          manual_BC_stretch; /* manual BC stretch. change #2 - harry */

{

  char  *calloc();
  TOKENPTR tokenptr = (TOKENTYPE *) calloc(1,sizeof(TOKENTYPE));

  if (tokenptr != NULL)
  {
    tokenptr->token_type    = toktyp;
    tokenptr->point_size    = ptsize;
    tokenptr->font          = font;
    tokenptr->horizontal_pos= horizontal;
    tokenptr->vertical_pos  = vertical;
    tokenptr->lr            = fontdir;
    tokenptr->begining      = beg_wrd;
    tokenptr->ending        = end_wrd;
    tokenptr->fillers_num	= NOFILLERS;	/* initialization ... Johny */
    tokenptr->manual_stretch_width = manual_stretch; /* change #1 - */
                       						   /* harry */
    tokenptr->stretch_width = NOSTRETCH;     /* default is no */
                    		               /* stretch. change #1 - */
                    			          /* harry */
    tokenptr->stretch_filler_width = NOSTRETCH;     /* default is no */
                                                    /* stretch. change #2 */
    tokenptr->manual_AC_width = manual_AC_stretch;  /* change #2 */
    tokenptr->manual_BC_width = manual_BC_stretch;  /* change #2 */
    tokenptr->AC_width = NOSTRETCH;                 /* default is no */
                                                    /* stretch. change #2 */
    tokenptr->BC_width = NOSTRETCH;                 /* default is no */
                                                    /* stretch. change #2 */
    tokenptr->char1         = char1;
    tokenptr->char2         = char2;
    tokenptr->char3         = char3;
    strcpy(tokenptr->font_name,f_name);
    tokenptr->width         = width;
    tokenptr->next          = NULL;
  }
  else
  {
    out_of_memory();
  }
 return (tokenptr);
}


/************************************************************************/
/************************************************************************/


/*
    this routine returns the basic (before stretch) width of a token 
                                                     */

int tokenBasicWidth (tokenptr)
TOKENPTR tokenptr;
{
	return tokenptr->width + tokenptr->manual_stretch_width + 
	  tokenptr->manual_AC_width + tokenptr->manual_BC_width; /* change #2 */
}


/************************************************************************/
/************************************************************************/


/*
    this routine returns the full (including stretch) width of a token 
                                                     */

int tokenFullWidth (tokenptr)
TOKENPTR tokenptr;
{
	return tokenBasicWidth(tokenptr) + tokenptr->stretch_width + 
	  tokenptr->stretch_filler_width + tokenptr->AC_width +     /* change #2 */
          tokenptr->BC_width;
}


/************************************************************************/
/************************************************************************/


/*
    this routine returns the full width of a word of tokens - change #2
                                                     */

int wordWidth (start)
TOKENPTR  start;                  /* ptr to start of word */
{
	TOKENPTR tokenptr = start;
	int width = 0;
	int finished = 0;

	while ((tokenptr != NULL) && !finished) {
		width += tokenFullWidth(tokenptr);
		if (tokenptr->ending)
		  finished = 1;
		tokenptr = tokenptr->next;
	}

	return width;
}

/************************************************************************/
/************************************************************************/


/*
    this routine returns the total letter stretch (including manual) of a token 
                                                     */

int tokenStretch (tokenptr)
TOKENPTR tokenptr;
{
	return tokenptr->stretch_width + tokenptr->manual_stretch_width;
}

/************************************************************************/
/************************************************************************/

/* The following 3 functions are change #2 */

/*
    this routine returns the total filler stretch of a token 
                                                     */

int tokenFillerStretch (tokenptr)
TOKENPTR tokenptr;
{
	return tokenptr->stretch_filler_width;
}

/************************************************************************/
/************************************************************************/


/*
    this routine returns the total letter AC stretch (including manual) of a token 
                                                     */

int tokenACStretch (tokenptr)
TOKENPTR tokenptr;
{
	return tokenptr->AC_width + tokenptr->manual_AC_width;
}

/************************************************************************/
/************************************************************************/


/*
    this routine returns the total letter BC stretch (including manual) of a token 
                                                     */

int tokenBCStretch (tokenptr)
TOKENPTR tokenptr;
{
	return tokenptr->BC_width + tokenptr->manual_BC_width;
}


/************************************************************************/
/************************************************************************/


/*
    this routine adds a token to the end of a line 
                                                     */

add_token (tokenptr,start,end)

  TOKENPTR     tokenptr;           /* token to be added        */
  TOKENPTR     *start;             /* ptr to start of line ptr */
  TOKENPTR     *end;               /* ptr to end of line ptr   */

{
  if (*start == NULL)
    *start = tokenptr;
  else {
    (*end)->next = tokenptr;
	tokenptr->next = NULL;        /* better to make sure it's NULL */
  }
  *end = tokenptr;
  return;
}

/************************************************************************/
/************************************************************************/

/*
    this routine adds a token to the end of a line without changing
    the tokenptr
                                                     */

simple_add_token (tokenptr,start,end)

  TOKENPTR     tokenptr;           /* token to be added        */
  TOKENPTR     *start;             /* ptr to start of line ptr */
  TOKENPTR     *end;               /* ptr to end of line ptr   */

{
  if (*start == NULL)
    *start = tokenptr;
  else (*end)->next = tokenptr;
  *end = tokenptr;
  return;
}

/************************************************************************/
/************************************************************************/

/*
    this routine pushes a token onto the front of a line 
                                                           */
push_token (tokenptr,start,end)

  TOKENPTR     tokenptr;            /* token to be push         */
  TOKENPTR     *start;              /* ptr to start of line ptr */
  TOKENPTR     *end;                /* ptr to end of line ptr   */

{

  if (*end == NULL) {
									/* reset the prev and next pointers ...
									   it's not sure that we call it ONLY
									   after new_token.  Johny  */
	(tokenptr)->next = NULL;
    *end = tokenptr;
  }
  else tokenptr->next = (*start);
  *start = tokenptr;
  return;
}

/*******************************************
********************************************
**                                        **
** this routine outputs an internal token ** 
**                                        **
********************************************
*******************************************/

put_token (tokenptr,start_of_line)

  TOKENPTR tokenptr;           /* token to output */
  bool     start_of_line;      /* indicates first character in line */

{
	static float current_slant = 0.0; /* current font slant if */
               				    /* in a slanted word */
	static int last_vertical_pos;     /* last slant vertical pos */
	int word_length;                  /* for slanted words */
	int vertical_offset;              /* for slanted words */

  if (start_of_line)
  {
    current_slant = 0.0;      /* change #2 */
    printf("H%d\n",tokenptr->horizontal_pos);
  }
 
  if (tokenptr->point_size != out_size)
  {
    printf ("s%d\n",tokenptr->point_size);
    out_size = tokenptr->point_size;
  }
  if (strcmp(tokenptr->font_name,out_font_name))
  {
    if (strcmp(tokenptr->font_name,out_fontable[tokenptr->font].name))
    {
      printf("x font %d %s\n",tokenptr->font,tokenptr->font_name);
      new_font(tokenptr->font,tokenptr->font_name,tokenptr->lr,
               out_fontable);
    }
    printf ("f%d\n",tokenptr->font);
    out_font = tokenptr->font;
    strcpy(out_font_name,tokenptr->font_name);
  } 

     /* all references to slant are change #2 */

     if ((font_slant(tokenptr->font) == 0.0) &&
	    (current_slant > 0)) {
		fprintf(stderr,"FFORTID: slant change in word\n");
		exit(1);
	}

	if (font_slant(tokenptr->font) > 0.0) {
		if (tokenptr->begining == BEGINING) {
			current_slant = font_slant(tokenptr->font);
			word_length = wordWidth(tokenptr);
			vertical_offset =
/*			  tan((double) current_slant*3.1415927/180)* */
			  0.40402623*
			    (word_length / 2.0);
			printf ("V%d\n",tokenptr->vertical_pos +
				vertical_offset);
			out_vertical = tokenptr->vertical_pos;
			last_vertical_pos = tokenptr->vertical_pos;
		} else if (font_slant(tokenptr->font) !=
			   current_slant) {
			fprintf(stderr,"FFORTID: slant change in word\n");
			exit(1); 
		}
	}

  if (tokenptr->vertical_pos != out_vertical)
  {
    printf ("V%d\n",tokenptr->vertical_pos);
    out_vertical = tokenptr->vertical_pos;
  }

  if (!start_of_line)
  {
    if (tokenptr->horizontal_pos < out_horizontal)
    {
      printf("H%d\n",tokenptr->horizontal_pos);
    }
    else if (tokenptr->horizontal_pos > out_horizontal) {
      int j =(tokenptr->horizontal_pos - out_horizontal);
      if (j < 100 && j > 9 && tokenptr->token_type == c_token){
              printf( "%d%c", j, tokenptr->char1);
              goto compress;
          }
      printf("h%d",(tokenptr->horizontal_pos - out_horizontal));
    }
  }


/* the following lines have changes to handle letter stretching. */
/* Change #1,#2 - harry */

  if (tokenStretch(tokenptr) > 0)
    printf("x X stretch %d\n",tokenStretch(tokenptr)); 
  if (tokenACStretch(tokenptr) > 0)
    printf("x X ACstretch %d\n",tokenACStretch(tokenptr)); 
  if (tokenBCStretch(tokenptr) > 0)
    printf("x X BCstretch %d\n",tokenBCStretch(tokenptr)); 
  switch (tokenptr->token_type) {
	case c_token: 
	  printf("c%c\n",tokenptr->char1); 
	  break;
	case C_token:
	  printf("C%c%c\n",tokenptr->char1,tokenptr->char2); 
	  break;
	case N_token:
	  printf("N%c%c%c\n",tokenptr->char1,tokenptr->char2,
		 tokenptr->char3); 
	  break;
  }
compress: /* thanks to Mulli Bahr hnncx compressed to nnx */
  out_horizontal = tokenptr->horizontal_pos;
	
	if (tokenFillerStretch(tokenptr) > 0) {
		printf("h%d",tokenFullWidth(tokenptr) - 
		       tokenFillerStretch(tokenptr));
		out_horizontal += tokenFullWidth(tokenptr) -  
		       tokenFillerStretch(tokenptr);
		printf("x X stretch %d\nChy\n",tokenFillerStretch(tokenptr)); 
	}

  /* added to insert the fillers ... Johny */
  if (tokenptr->fillers_num > 0) {
	int i;
	printf("h%d",tokenFullWidth(tokenptr));
	printf("Chy\n");
	for (i=0; i<(tokenptr->fillers_num)-1; i++)
		printf("h%dChy\n",tokenptr->filler_width);
	printf("h%d",tokenptr->filler_width);
	out_horizontal += tokenptr->fillers_num * tokenptr->filler_width +
	  tokenFullWidth(tokenptr);
  }

  if (tokenptr->ending == END) {
	  if (font_slant(tokenptr->font) > 0.0) {
		  printf ("V%d\n",last_vertical_pos); 
		  current_slant = 0.0; 
	  }
	  printf("w");
}
	
  return;
}


/*************************************************************
**************************************************************
**                                                          **
** this routine outputs a new page token and its associated **
** font and point size tokens and resets its font and size  **
** to null and 0 so as to force output of f and s commands  **
**                                                          **
**************************************************************
*************************************************************/

put_page_token(page_token)

  char  *page_token;          /* lex page token value */

{

  extern char  yytext[];     /* lex token value */
  
  printf("V%d\n",in_vertical); 
				/* ditroff guarantees and thus drivers
				require that before page_token, a V0 token! 
				Mulli Bahr, Oct, 86*/
  out_vertical = in_vertical;
  printf("%s\n",page_token);

  out_size = 0;                            /* reset out_size to zero ... */
  out_font_name[0] = '\0';                 /* reset out_font_name to empty
         string  so that the next token print will be forced to dump out s and f
         commands. On some devices the current size and page are not remembered
         across page boundaries, especially if there's a bunch of "x font XX"s.
         Besides the ditroff output always seems to have the s and f issued just
         before the text on each page */

  return;
}


/***********************************************************************/
/***********************************************************************/


/*
    this routine deallocates a line of tokens
                                                */

free_line(start,end)

  TOKENPTR  *start;           /* ptr to start of line */
  TOKENPTR  *end;             /* ptr to end of line   */

{
  TOKENPTR  tokenptr;         /* next element in line */

  while( (tokenptr = *start) != NULL)
  {
    *start = tokenptr->next;
    cfree(tokenptr);
  }
  *end = NULL;
  return;
}

/************************************************************************/ 
/************************************************************************/ 
