/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */

#include	<stdio.h>
#include "macros.h"

typedef struct {
	int	space_width;
	int	no_width_entries;
	char	is_special_font;
	int  stretchability;          /* change #2 */
	float slant;
     int   connection_stretch;
	char	font_name[10];
} Fontinfo;

#define	MAXNOFONTS	255
#define	MAXWIDENTRIES	256
#define	NOCHARSINBIGGESTFONT	MAXWIDENTRIES-1	/* for no biggestfont in DESC */
#define	MAXNOCHARS	512	/* characters with two-letter or --- names */
#define	SIZECHARINDXTABLE	(MAXNOCHARS + 128-32)	/* includes ascii chars,
				but not non-graphics */

Fontinfo	basic_font_info[MAXNOFONTS+1];
char	font_name[MAXNOFONTS][10];

int	no_of_fonts;
int	indx_1st_spec_font;	/* index of first special font */
int	size_char_table;
int	unit_width;
int	units_per_inch;
int	no_chars_in_biggest_font = NOCHARSINBIGGESTFONT;

int	size_char_name;
char	char_name[5*MAXNOCHARS];	/* 2 or 3 letter character names,
					          including \0 for each */
short	char_table[MAXNOCHARS+1];	/* index of characters in char_name */

char	*char_indx_table[MAXNOFONTS+1];    /* fitab*/
char	*code_table[MAXNOFONTS+1];         /* codetab*/
char	*width_table[MAXNOFONTS+1];	     /* widtab would be a better name */

char *connect_table[MAXNOFONTS+1];     /* connectivity table. Change */
				                   /* #1 - harry */
char *stretch_table[MAXNOFONTS+1];     /* stretchability table. Change */
				                   /* #1  - harry */

#define	FATAL	1
#define	BYTEMASK	0377

/* char	*fontdir	= "/usr/lib/font"; */
char	*fontdir	= "/home/harryh/usr/lib/font";

extern   char	*device;	/* output device */

/**************************************************************
***************************************************************
**                                                           **
** this routine initializes the device and font width tables **
**                                                           **
***************************************************************
**************************************************************/

width_init()	/* read in font and code files, etc. */
{
	int i, j;
	char *malloc();
	char temp[60];
	FILE *descfile;
	char word[100], *ptr;

	sprintf(temp, "%s/dev%s/DESC", fontdir, device);
	if ((descfile = fopen(temp, "r")) == NULL){
		error(FATAL, "can't open DESC for %s\n", temp);
	}
	while (fscanf(descfile, "%s", word) != EOF) {
		if (strcmp(word, "res") == 0) {
			fscanf(descfile, "%d", &units_per_inch);
		} else if (strcmp(word, "unitwidth") == 0) {
			fscanf(descfile, "%d", &unit_width);
		} else if (strcmp(word, "fonts") == 0) {
			fscanf(descfile, "%d", &no_of_fonts);
			if (no_of_fonts > MAXNOFONTS) {
				error(FATAL,
				"have more fonts, %d, than the maximum allowed, %d\n",
				 no_of_fonts, MAXNOFONTS);
			}
			/* start at 1 to leave 0 for default font */
			for (i = 1; i <= no_of_fonts; i++) {
				fscanf(descfile, "%s", font_name[i]);
			}
		} else if (strcmp(word, "biggestfont") == 0) {
			fscanf(descfile, "%d", &no_chars_in_biggest_font);
			if (no_chars_in_biggest_font > NOCHARSINBIGGESTFONT) {
				error(FATAL,
					"biggestfont of DESC too big, %d\n",
					no_chars_in_biggest_font);
			}
		} else if (strcmp(word, "charset") == 0) {
			ptr = char_name;
			size_char_table = 0;
			while (fscanf(descfile, "%s", ptr) != EOF) {
				if (size_char_table == MAXNOCHARS-1) {
					error(FATAL,
					"have more chars than the maximum allowed, %d\n",
				 	MAXNOCHARS);
				}
				char_table[size_char_table++] = ptr - char_name;
				while (*ptr++)	/* skip to end of char name */
					;
			}
			size_char_name = ptr - char_name;
			char_table[size_char_table++] = 0; /* end with \0 */
		} else
		/* skip anything else */
		while (getc(descfile) != '\n');
	}
	fclose(descfile);

	width_table[0] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
	code_table[0] =(char *) malloc(MAXWIDENTRIES * sizeof(char));

	/* Change #1 - harry */

	connect_table[0] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
	stretch_table[0] =(char *) malloc(MAXWIDENTRIES *
					  sizeof(char));
	for (j = 0; j <= MAXWIDENTRIES-1; j++) {
		width_table[0][j] = 0;
		code_table[0][j] = 0;

		/* Change #1 - harry */

		connect_table[0][j] = 0;
		stretch_table[0][j] = 0;
	}
	char_indx_table[0] =(char *) malloc(SIZECHARINDXTABLE * sizeof(char));
	for (j = 0; j <= SIZECHARINDXTABLE-1; j++) {
		char_indx_table[0][j] = 0;
	}
	basic_font_info[0].space_width = 0;
	basic_font_info[0].no_width_entries = no_chars_in_biggest_font+1;
	basic_font_info[0].is_special_font = 0;
	basic_font_info[0].stretchability = FONT_NO_STRETCH;   /* change #2 */
	basic_font_info[0].slant = 0;
	basic_font_info[0].connection_stretch = FIXED_FILLER;
	basic_font_info[0].font_name[0] = '\0';


/*	deviceprint();	/* debugging print of device */
/*	fontprint(0); /*debugging print of font tables [0]*/

	for (i = 1; i <= no_of_fonts; i++) {
		width_table[i] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
		code_table[i] =(char *) malloc(MAXWIDENTRIES * sizeof(char));

		/* Change #1 - harry */

		connect_table[i] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
		stretch_table[i] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
		for (j = 0; j <= MAXWIDENTRIES-1; j++) {
			width_table[i][j] = 0;
			code_table[i][j] = 0;

			/* Change #1 - harry */

			connect_table[i][j] = 0;
			stretch_table[i][j] = 0;
		}
		char_indx_table[i] =(char *) malloc(SIZECHARINDXTABLE *
					sizeof(char));
		for (j = 0; j <= SIZECHARINDXTABLE-1; j++) {
			char_indx_table[i][j] = 0;
		}
		getfontinfo(font_name[i],i); 
/*		fontprint(i); /*debugging print of font tables [i]*/
	}
}

getfontinfo(font_name,pos)
char *font_name;
int pos;
{
	FILE *fontfile;
	int i, no_width_entries, space_width;
	float slant;   /* change #2 */
	int c;
	char buffer[100], word[30],
		char_c[10], wid[10], asc_des[10], code[10],
	        connect_c[10], stretch_c[10];	/* Change #1 - harry */

     /* chnage #2 */

	basic_font_info[pos].space_width = 0;
	basic_font_info[pos].no_width_entries = 0;
	basic_font_info[pos].is_special_font = 0;
	basic_font_info[pos].stretchability = FONT_NO_STRETCH;
	basic_font_info[pos].slant = 0.0;
	basic_font_info[pos].connection_stretch = FIXED_FILLER;
	basic_font_info[pos].font_name[0] = '\0';

	sprintf(buffer, "%s/dev%s/%s", fontdir, device, font_name);
	if ((fontfile = fopen(buffer, "r")) == NULL){
		error(FATAL, "can't open width table for %s\n", font_name);
	}
	while (fscanf(fontfile, "%s", word) != EOF) {
		if (word[0] == '#')
			while(getc(fontfile) != '\n');
		else if (strcmp(word, "name") == 0)
			fscanf(fontfile, "%s", basic_font_info[pos].font_name);
		else if (strcmp(word, "special") == 0)
			basic_font_info[pos].is_special_font = 1;
		else if (strcmp(word, "spacewidth") == 0) {
			fscanf(fontfile, "%d",&space_width);
			basic_font_info[pos].space_width = space_width;
			if (space_width == 0) {
				/* Rounding error fix. Change #1 - harry */

				width_table[pos][0] = units_per_inch * unit_width / 72.0 / 3.0;
			} else {
				width_table[pos][0] = space_width;
			}
                    /* change #2 */

		} else if (strcmp(word, "stretchable:") == 0) {
			while ((c = getc(fontfile)) != '\n') {
				if (c == 'l') {
					basic_font_info[pos].stretchability |= FONT_LETT_STRETCH;
				} else if (c == 'c')
				  basic_font_info[pos].stretchability |= FONT_CONN_STRETCH;
			}
			ungetc('\n', fontfile);
		} else if (strcmp(word, "slant") == 0) {
			fscanf(fontfile, "%f",&slant);
			if (slant < 0.0)
			  slant = 0.0;
			basic_font_info[pos].slant = slant;
		} else if (strcmp(word, "connection") == 0) {
			fscanf(fontfile, "%s", word);
			if (strcmp(word, "stretch:") != 0)
			  continue;
			fscanf(fontfile, "%s", word);
			if (strcmp(word, "fixed") == 0) {
				basic_font_info[pos].connection_stretch = 
				  FIXED_FILLER;
				fscanf(fontfile, "%s", word);
			} else if (strcmp(word, "stretchable") == 0) { 
				fscanf(fontfile, "%s", word);
				if (strcmp(word, "filler") == 0) {
					basic_font_info[pos].connection_stretch =
					  STRETCHABLE_FILLER;
				} else if (strcmp(word, "connections") == 0)
				  basic_font_info[pos].connection_stretch =
				    STRETCHABLE_CONNECTIONS;
			}
		} else if (strcmp(word, "charset") == 0) {
			while(getc(fontfile) != '\n');
			no_width_entries = 0;
			/* widths are origin 1 so char_indx_table entry of
			0 can mean not there */
			while (fgets(buffer, 100, fontfile) != NULL) {

				/* Change #1 - harry */

				sscanf(buffer, "%s %s %s %s %s %s",
					char_c, wid, asc_des, code,
				       connect_c, stretch_c);

				if (wid[0] != '"') { /* not a ditto */
					no_width_entries++;
					width_table[pos][no_width_entries] =
						atoi(wid);

					/* change #2 */

					connect_table[pos][no_width_entries] = connect_c[0];
					stretch_table[pos][no_width_entries] = stretch_c[0];

					if (code[0] == '0')
						sscanf(code, "%o", &i);
					else
						sscanf(code, "%d", &i);
					code_table[pos][no_width_entries] = i;
				}
				/* otherwise a synonym for previous character,
				so leave previous values intact */
				if (strlen(char_c) == 1)	/* it's ascii */
					char_indx_table[pos][char_c[0]-32] =
						no_width_entries;
						/* char_indx_table origin
						omits non-graphics */
				else if (strcmp(char_c, "---") != 0) {
						/* it has a 2-char name */
					for (i = 0; i <= size_char_table; i++)
						if (strcmp(
						    &char_name[char_table[i]],
						    char_c) == 0) {
							char_indx_table[pos][i+128-32] =
								no_width_entries;
								/* starts after the ascii */
							break;
						}
					if (i >= size_char_table)
						error(FATAL,
						    "font %s: %s not in charset\n",
						    font_name, char_c);
				}
			}
			if (no_width_entries > MAXWIDENTRIES) {
				error(FATAL,
					"font has %d characters, too big\n",
					no_width_entries);
			}
			basic_font_info[pos].no_width_entries =
				no_chars_in_biggest_font+1;
		}
		  else while(getc(fontfile) != '\n');
	}
	fclose(fontfile);
}

/*******************************************************
********************************************************
**                                                    **
** this routine prints the specified font width table **
**                                                    **
********************************************************
*******************************************************/

fontprint(i)	/* debugging print of font i (0,...) */
{
	int jj, kk, nn;

	printf("font %d:\n", i);
	nn = basic_font_info[i].no_width_entries;

	printf("base=0xxxxxx, nchars=%d, spec=%d, name=%s, width_table=0xxxxxx, char_indx_table=0xxxxxx, code_table=0xxxxxx\n",
		nn, basic_font_info[i].is_special_font, basic_font_info[i].font_name);

	printf("\nwidths:\n");
	for (jj=0; jj <= nn; jj++) {
		printf(" %2d", width_table[i][jj] & BYTEMASK);
		if (jj % 20 == 19) printf("\n");
	}

	printf("\nchar_indx_table:\n");
	for (jj=0; jj < size_char_table + 128-32; jj++) {
		printf(" %2d", char_indx_table[i][jj] & BYTEMASK);
		if (jj % 20 == 19) printf("\n");
	}

	printf("\ncode_table:\n");
	for (jj=0; jj <= nn; jj++) {
		printf(" %2d", code_table[i][jj] & BYTEMASK);
		if (jj % 20 == 19) printf("\n");
	}

	printf("\n");
}

/*******************************************************
********************************************************
**                                                    **
** this routine prints the device table **
**                                                    **
********************************************************
*******************************************************/

deviceprint()	/* debugging print of device */
{
	int j;
	int jj;

	printf("device:\n");

	printf("res=%d nfonts=%d nchtab=%d unitwidth=%d lchname=%d",
		units_per_inch, no_of_fonts, size_char_table, unit_width,
		size_char_name);

	printf("\nchtab:\n");
	for (jj=0; jj <= size_char_table-1; jj++) {
		printf(" %2d", char_table[jj]);
		if (jj % 20 == 19) printf("\n");
	}

	printf("\nchname:\n");
	for (jj=0; jj <= size_char_table-1; jj++) {
		printf(" %s", &char_name[char_table[jj]]);
		if (jj % 20 == 19) printf("\n");
	}

	printf("\n");
}

/******************************************************
*******************************************************
**                                                   **
** this routine loads the specified font width table **
**                                                   **
*******************************************************
******************************************************/

loadfont(pos, font_name) /*load info for font font_name on position pos (0...)*/
int pos;
char *font_name;
{
	int j;

	if (pos < 0 || pos > MAXNOFONTS)
		error(FATAL, "illegal fp command %d %s", pos, font_name);
	if (strcmp(font_name, basic_font_info[pos].font_name) == 0)
		return; /* if font font_name is already mounted on position pos
                           return immediately */
	width_table[pos] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
	code_table[pos] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
	connect_table[pos] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
	stretch_table[pos] =(char *) malloc(MAXWIDENTRIES * sizeof(char));
	for (j = 0; j <= MAXWIDENTRIES-1; j++) {
		width_table[pos][j] = 0;
		code_table[pos][j] = 0;
		connect_table[pos][j] = 0;
		stretch_table[pos][j] = 0;
	}
	char_indx_table[pos] =(char *) malloc(SIZECHARINDXTABLE *
				sizeof(char));
	for (j = 0; j <= SIZECHARINDXTABLE-1; j++) {
		char_indx_table[pos][j] = 0;
	}
	getfontinfo(font_name,pos); 
}


error(f, s, a1, a2, a3, a4, a5, a6, a7) {
	fprintf(stderr, "ffortid: ");
	fprintf(stderr, s, a1, a2, a3, a4, a5, a6, a7);
	fprintf(stderr, "\n");
	if (f)
          exit(1);
}


/***********************************************************************
************************************************************************
**                                                                    **
** this routine returns the index of funny character s in tables      **
**                                                                    **
************************************************************************
***********************************************************************/


funnyCharIndex(s)	/* s is a funny char name */
char *s;
{
	int i;

	for (i = 0; i < size_char_table; i++)
	     if (strcmp(&char_name[char_table[i]], s) == 0)
	          break;
	if (i < size_char_table)
          return(i+128);
     else return(0);
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine determines the index of abs char n in tables    **
**                                                              **
******************************************************************
*****************************************************************/

absCharIndex(n,in_font)	
int n;
int in_font;
{
	register int i, ncf;
	
	ncf= basic_font_info[in_font].no_width_entries & BYTEMASK;
	for (i=0; i< ncf; i++)
		if ((unsigned char)code_table[in_font][i] == n)
		/* a bug fix for the \N'xxx' to work, when xxx > 128 */
			return i;
	return 0;
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine converts width table value to goobies for a     **
**   certain point size                                         **
**                                                              **
******************************************************************
*****************************************************************/

int widthToGoobies(width, point_size)
int width;
int point_size;
{
	return (int) (((float) width * point_size / unit_width) + 0.5) ;
}


/***********************************************************************
************************************************************************
**                                                                    **
** this routine determines the width of the specified funny character **
**                                                                    **
************************************************************************
***********************************************************************/


width2(s,in_size,in_font)	/* s is a funny char name */
char *s;
int  in_size;		/* Johny */
int	 in_font;
{
     return(width1(funnyCharIndex(s),in_size,in_font));
}


/* All the following functions were changed or added in change #1 - */
/* harry */

/*****************************************************************
******************************************************************
**                                                              **
** this routine determines the width of the specified character **
**                                                              **
** in_size is passed as a parameter as this procedure is used   **
** to calculate the filler width for each word of the line.	    **
** it is necessary because the in_size at the start and end of  **
** the line may be different. Johny				         **
** 								                        **
******************************************************************
*****************************************************************/


width1(c,in_size,in_font)	/* output char c */
int c;
int in_size;
int in_font;
{
	char *pw;
	register char *p;
	register int i, k;
	int j, w, width;

	c -= 32;
	if (c <= 0)
		return(widthToGoobies(width_table[in_font][0],in_size));
	k = in_font;
	i = char_indx_table[in_font][c] & BYTEMASK;
	if (i != 0) {	/* it's on this font */
		pw = width_table[in_font];
	} else if (indx_1st_spec_font > 0) { /* on special (we hope) */
		for (k=indx_1st_spec_font, j=0;
				j < no_of_fonts; j++, k = k % no_of_fonts + 1)
			if ((i = char_indx_table[k][c] & BYTEMASK) != 0) {
				pw = width_table[k];
				break;
			}
	}
	if (i == 0 || j == no_of_fonts) {
		return(widthToGoobies(width_table[in_font][0],in_size));
	}
	width = pw[i] & BYTEMASK;

	width = widthToGoobies(width, in_size);
        return(width);
}



/*****************************************************************
******************************************************************
**                                                              **
** this routine determines the width of the character whose     **
** code is n, i.e. specified Nn                                 **
**                                                              **
******************************************************************
*****************************************************************/


widthn(pn,in_size,in_font)	/* output char with abs code *pn */
char *pn;
int in_size;
int in_font;
{
	char *pw;
	register int i;
	int n,width;
	
	sscanf(pn, "%d", &n); /*get the string *pn and convert to integer*/

	i= absCharIndex (n,in_font);
	pw=width_table[in_font];
	width = pw[i] & BYTEMASK;

	/* rounding error fix. Change #1 - harry */

	width = widthToGoobies(width, in_size);
	return(width);
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a funny character is a connect  **
**     previous letter                                          **
**                                                              **
******************************************************************
*****************************************************************/

int connectable2(s,in_font)
char* s;
int in_font;
{
     return(connectable1(funnyCharIndex(s),in_font));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a funny character is a connect  **
**     after letter                                             **
**                                                              **
******************************************************************
*****************************************************************/

int connectAfter2(s,in_font)
char* s;
int in_font;
{
     return(connectAfter1(funnyCharIndex(s),in_font));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a funny character is a          **
**   stretchable letter                                         **
**                                                              **
******************************************************************
*****************************************************************/

int stretchable2(s,in_font)
char* s;
int in_font;
{
     return(stretchable1(funnyCharIndex(s),in_font));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns the connectivity of a regular character **
**                                                              **
******************************************************************
*****************************************************************/

int connect1(c,in_font)
int c;
int in_font;
{
	char *pc;
	register char *p;
	register int i, k;
	int j, w;

	c -= 32;
	if (c <= 0) 
		return (connect_table[in_font][0]);
	k = in_font;
	i = char_indx_table[in_font][c] & BYTEMASK;
	if (i != 0) {	/* it's on this font */
		pc = connect_table[in_font];
	} else if (indx_1st_spec_font > 0) { /* on special (we hope) */
		for (k=indx_1st_spec_font, j=0;
				j < no_of_fonts; j++, k = k % no_of_fonts + 1)
			if ((i = char_indx_table[k][c]) != 0) {
				pc = connect_table[k];
				break;
			}
	}
	if (i == 0 || j == no_of_fonts) 
		return (connect_table[in_font][0]);

	return pc[i];
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a character is a connect        **
**     previous letter                                          **
**                                                              **
******************************************************************
*****************************************************************/

int connectable1(c,in_font)
int c;
int in_font;
{
	char ch;

     ch = connect1(c,in_font);
	return ((ch == CONNECTPREVIOUS) || (ch == CONNECTBOTH));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a character is a connect        **
**     after letter                                             **
**                                                              **
******************************************************************
*****************************************************************/

int connectAfter1(c,in_font)
int c;
int in_font;
{
	char ch;

     ch = connect1(c,in_font);
	return ((ch == CONNECTAFTER) || (ch == CONNECTBOTH));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a character is a stretchable    **
**   letter                                                     **
**                                                              **
******************************************************************
*****************************************************************/

int stretchable1(c,in_font)
int c;
int in_font;
{
	char *ps;
	register char *p;
	register int i, k;
	int j, w;

	c -= 32;
	if (c <= 0)
		return (stretch_table[in_font][0] == STRETCHABLE);
	k = in_font;
	i = char_indx_table[in_font][c] & BYTEMASK;
	if (i != 0) {	/* it's on this font */
		ps = stretch_table[in_font];
	} else if (indx_1st_spec_font > 0) { /* on special (we hope) */
		for (k=indx_1st_spec_font, j=0;
				j < no_of_fonts; j++, k = k % no_of_fonts + 1)
			if ((i = char_indx_table[k][c]) != 0) {
				ps = stretch_table[k];
				break;
			}
	}
	if (i == 0 || j == no_of_fonts) 
		return (stretch_table[in_font][0] == STRETCHABLE);

	return (ps[i] == STRETCHABLE);
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether abs char n is a connect         **
**     previous letter                                          **
**                                                              **
******************************************************************
*****************************************************************/

int connectablen(n,in_font)
int n;
int in_font;
{
	char c;

	c = connect_table[in_font][absCharIndex(n,in_font)];
	return ((c == CONNECTPREVIOUS) || (c == CONNECTBOTH));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether abs char n is a connect         **
**     after letter                                             **
**                                                              **
******************************************************************
*****************************************************************/

int connectAftern(n,in_font)
int n;
int in_font;
{
	char c;

	c = connect_table[in_font][absCharIndex(n,in_font)];
	return ((c == CONNECTAFTER) || (c == CONNECTBOTH));
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether abs char n is a stretchable     **
**   letter                                                     **
**                                                              **
******************************************************************
*****************************************************************/

int stretchablen(n,in_font)
int n;
int in_font;
{
	return (stretch_table[in_font][absCharIndex(n,in_font)] 
                                             == STRETCHABLE);
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a font has letter stretching    **
**                                                              **
******************************************************************
*****************************************************************/

int font_letter_stretchable(font)
int font;
{
	return (basic_font_info[font].stretchability & FONT_LETT_STRETCH);
}


/*****************************************************************
******************************************************************
**                                                              **
** this routine returns whether a font has connection stretching**
**                                                              **
******************************************************************
*****************************************************************/

int font_connection_stretchable(font)
int font;
{
	return (basic_font_info[font].stretchability & FONT_CONN_STRETCH);
}
 

/*****************************************************************
******************************************************************
**                                                              **
** this routine returns a fonts slant in degrees                **
**                                                              **
******************************************************************
*****************************************************************/

float font_slant(font)
int font;
{
        return (basic_font_info[font].slant);
}
 

/*****************************************************************
******************************************************************
**                                                              **
** this routine returns a fonts connection stretch type         **
**                                                              **
******************************************************************
*****************************************************************/

int font_connection_stretch(font)
int font;
{
        return (basic_font_info[font].connection_stretch);
}
